from . constants import *
import datetime
import acrobind.protection
import acrort


def week_number(date, week_beginning=1):
    return (date - datetime.timedelta(days=week_beginning - 1)).isocalendar()[1]


def calculate_simple_backup_set(slice, predecessor, schedule_type, week_beginning=1):
    if predecessor is None:
        return BACKUP_SET_MONTHLY
    slice_time = slice['creation_time']
    predecessor_time = predecessor['creation_time']
    if slice_time <= predecessor_time:
        acrort.common.make_logic_error('slice_time={} while predecessor_time={}'.format(slice_time, predecessor_time)).throw()
    if schedule_type == 'monthly' or slice_time.year != predecessor_time.year or slice_time.month != predecessor_time.month:
        return BACKUP_SET_MONTHLY
    if schedule_type == 'weekly' or slice_time.day != predecessor_time.day and week_number(slice_time, week_beginning) != week_number(predecessor_time, week_beginning):
        return BACKUP_SET_WEEKLY
    if schedule_type == 'daily' or predecessor_time.day != slice_time.day:
        return BACKUP_SET_DAILY
    return BACKUP_SET_HOURLY


def calculate_weekly_daily_backup_set(slice, predecessor, week_beginning=1):
    backup_set = calculate_simple_backup_set(slice, predecessor, 'daily', week_beginning)
    if backup_set == BACKUP_SET_HOURLY:
        return BACKUP_SET_DAILY
    if backup_set == BACKUP_SET_MONTHLY:
        return BACKUP_SET_WEEKLY
    return backup_set


def calculate_monthly_weekly_daily_backup_set(slice, predecessor, week_beginning=1):
    backup_set = calculate_simple_backup_set(slice, predecessor, 'daily', week_beginning)
    if backup_set == BACKUP_SET_HOURLY:
        return BACKUP_SET_DAILY
    return backup_set


def get_schedule_type(backup_plan):
    schedule_types = {
        SCHEDULE_TYPE_NOT_SPECIFIED: None,
        SCHEDULE_TYPE_HOURLY: 'hourly',
        SCHEDULE_TYPE_DAILY: 'daily',
        SCHEDULE_TYPE_WEEKLY: 'weekly',
        SCHEDULE_TYPE_MONTHLY: 'monthly'
    }
    single_schedule_type = lambda: schedule_types[backup_plan.Scheme.Parameters.BackupSchedule.Schedule.Type.ref]
    scheme_schedule_types = {
        acrobind.protection.BACKUP_SCHEME_CUSTOM: (lambda: 'custom'),
        acrobind.protection.BACKUP_SCHEME_WEEKLY_FULL_DAILY_INCR: (lambda: 'weekly-daily'),
        acrobind.protection.BACKUP_SCHEME_MONTHLY_FULL_WEEKLY_DIFF_DAILY_INCR: (lambda: 'monthly-weekly-daily'),
        acrobind.protection.BACKUP_SCHEME_ALWAYS_FULL: single_schedule_type,
        acrobind.protection.BACKUP_SCHEME_ALWAYS_INCREMENTAL: single_schedule_type,
        acrobind.protection.BACKUP_SCHEME_SIMPLE: single_schedule_type,
        acrobind.protection.BACKUP_SCHEME_GFS: (lambda: 'gfs')
    }
    get_schedule_type = scheme_schedule_types[backup_plan.Scheme.Type.ref]
    return get_schedule_type()


def create_backup_set_calculator(schedule_type, week_beginning=1):
    kind_mapping = {
        acrobind.protection.BACKUP_TYPE_FULL: BACKUP_SET_FULL,
        acrobind.protection.BACKUP_TYPE_DIFFERENTIAL: BACKUP_SET_DIFF,
        acrobind.protection.BACKUP_TYPE_INCREMENTAL: BACKUP_SET_INC
    }
    calculate_by_kind = lambda slice, predecessor: kind_mapping[slice['kind']]
    calculators = {
        'monthly': (lambda slice, predecessor: calculate_simple_backup_set(slice, predecessor, 'monthly', week_beginning)),
        'weekly': (lambda slice, predecessor: calculate_simple_backup_set(slice, predecessor, 'weekly', week_beginning)),
        'daily': (lambda slice, predecessor: calculate_simple_backup_set(slice, predecessor, 'daily', week_beginning)),
        'hourly': (lambda slice, predecessor: calculate_simple_backup_set(slice, predecessor, 'hourly', week_beginning)),
        'weekly-daily': (lambda slice, predecessor: calculate_weekly_daily_backup_set(slice, predecessor, week_beginning)),
        'monthly-weekly-daily': (lambda slice, predecessor: calculate_monthly_weekly_daily_backup_set(slice, predecessor, week_beginning)),
        'custom': calculate_by_kind,
        'gfs': calculate_by_kind
    }
    if schedule_type in calculators:
        return calculators[schedule_type]
    return calculate_by_kind


def get_week_beginning(backup_plan):
    return 1 + (backup_plan.Options.BackupOptions.Additional.WeeklyBackupDay.ref + 6) % 7