import acrort
import datetime
from . command_ids import *
from . constants import *
from . backup_set import *
from . retention import *
from . replication import *
from . slices import *


def make_command_spec(command_id, argument, min_progress, max_progress):
    return {
        'CommandID': ('guid', command_id),
        'Argument': argument,
        'MinProgress': ('dword', min_progress),
        'MaxProgress': ('dword', max_progress)
    }


def make_workflow(commands):
    result = {'Commands': commands}
    return acrort.plain.Unit(result)


def request_for_make_shallow_copy(source, target):
    argument = {
        'ArchiveName': source['archive_name'],
        'SourceLocation': {
            'Address': source['location_uri'],
            'Credentials': source['location_credentials']
        },
        'TargetLocation': {
            'Address': target['location_uri'],
            'Credentials': target['location_credentials']
        }
    }
    return (DMS_MAKE_SHALLOW_COPY_COMMAND_ID, argument)


def get_max_chain_length(slicing):
    return slicing.get('MaxValue', ('dword', 0))


def request_for_replicate_slices(source, slice_keys, target):
    argument = {
        'SourceLocationUri': source['location_uri'],
        'SourceLocationLogon': source['location_credentials'],
        'SourceArchiveName': source['archive_name'],
        'ArchivePassword': source['archive_password'],
        'SourceSliceKeys': acrort.plain.Unit(slice_keys),
        'TargetLocationUri': target['location_uri'],
        'TargetLocationLogon': target['location_credentials'],
        'TargetArchiveName': target['archive_name'],
        'TargetBackupChainLength': get_max_chain_length(target['slicing'])
    }

    return (DMS_REPLICATE_BACKUP_COMMAND_ID, argument)


def request_for_delete_slices(stage, slice_uris):
    argument = {
        'DeleteSlicesTargets': slice_uris,
        'ArchivePassword': stage['archive_password'],
        'LocationCredentials': stage['location_credentials']
    }
    return (DMS_DELECT_SLICES_COMMAND_ID, argument)


def calculate_just_cleanup_workflow(just_cleanup_rules, backup_set_calculator, cleanup_time, source_slice_selector, source_stage):
    just_cleanup_by_time_rules = [item for item in just_cleanup_rules if item.BackupCountUpperLimit.ref == 0 and item.DeleteOlderBackupsTimePeriod.Value.ref > 0]
    just_cleanup_by_number_rules = [item for item in just_cleanup_rules if item.BackupCountUpperLimit.ref > 0 and item.DeleteOlderBackupsTimePeriod.Value.ref == 0]
    source_slices = select_slices_on_source(source_slice_selector)
    if just_cleanup_by_time_rules:
        retention_predicate = create_retention_predicate(just_cleanup_by_time_rules, backup_set_calculator, cleanup_time)
        slices = select_slices_abandoned_by_time(source_slices, retention_predicate)
    elif just_cleanup_by_number_rules:
        slices = select_slices_abandoned_by_number(source_slices, backup_set_calculator, just_cleanup_by_number_rules)
    if len(slices) == 0:
        return []

    slice_uris = [slice['uri'] for slice in slices]
    return [request_for_delete_slices(source_stage, preserve_last_slice(source_slices, slice_uris))]


def calculate_copy_on_appearance_workflow(copy_on_appearance_rules, backup_set_calculator, source_slice_selector, target_slice_selector, source_stage, target_stage):
    replication_predicate = create_replication_predicate(copy_on_appearance_rules, backup_set_calculator)
    create_archive, slices = select_slices_to_replicate(source_slice_selector, target_slice_selector, replication_predicate)
    if len(slices) == 0:
        return []
    slice_keys = [slice['id'] for slice in slices]
    requests = []
    if create_archive:
        requests = [request_for_make_shallow_copy(source_stage, target_stage)]
    return requests + [request_for_replicate_slices(source_stage, slice_keys, target_stage)]


def calculate_move_on_cleanup_workflow(service, move_on_cleanup_rules, backup_set_calculator, cleanup_time, source_slice_selector, source_stage, target_stage):
    move_on_cleanup_by_time_rules = [item for item in move_on_cleanup_rules if item.BackupCountUpperLimit.ref == 0 and item.DeleteOlderBackupsTimePeriod.Value.ref > 0]
    move_on_cleanup_by_number_rules = [item for item in move_on_cleanup_rules if item.BackupCountUpperLimit.ref > 0 and item.DeleteOlderBackupsTimePeriod.Value.ref == 0]
    source_slices = select_slices_on_source(source_slice_selector)
    if move_on_cleanup_by_time_rules:
        retention_predicate = create_retention_predicate(move_on_cleanup_by_time_rules, backup_set_calculator, cleanup_time)
        slices = select_slices_abandoned_by_time(source_slices, retention_predicate)
    elif move_on_cleanup_by_number_rules:
        slices = select_slices_abandoned_by_number(source_slices, backup_set_calculator, move_on_cleanup_by_number_rules)
    if len(slices) == 0:
        return []
    create_archive = select_slices(service, target_stage) is None
    slice_keys = [slice['id'] for slice in slices]
    slice_uris = [slice['uri'] for slice in slices]
    requests = []
    if create_archive:
        requests = [request_for_make_shallow_copy(source_stage, target_stage)]
    replicate_request = request_for_replicate_slices(source_stage, slice_keys, target_stage)
    cleanup_request = request_for_delete_slices(source_stage, preserve_last_slice(source_slices,slice_uris))
    return requests + [replicate_request, cleanup_request]


def calculate_workflow(service, backup_plan, stage_index):
    stages = [make_stage_spec(backup_plan, stage) for _, stage in backup_plan.Route.Stages]
    stage_count = len(stages)
    if stage_index >= stage_count:
        acrort.common.make_logic_error('stage_index={} while stage_count={}'.format(stage_index, stage_count)).throw()

    cleanup_time = datetime.datetime.now()
    source_stage = stages[stage_index]
    schedule_type = get_schedule_type(backup_plan)
    week_start = get_week_beginning(backup_plan)
    backup_set_calculator = create_backup_set_calculator(schedule_type, week_start)
    source_slice_selector = lambda: select_slices(service, source_stage)
    after_backup_rules = [item for _, item in source_stage['rules'] if item.AfterBackup.ref]
    just_cleanup_rules = [item for item in after_backup_rules if item.StagingOperationType.ref == STAGING_TYPE_JUST_CLEANUP]
    cleanup_workflow = []
    if just_cleanup_rules:
        cleanup_workflow = calculate_just_cleanup_workflow(just_cleanup_rules, backup_set_calculator, cleanup_time, source_slice_selector, source_stage)

    if stage_index + 1 >= stage_count:
        return cleanup_workflow

    target_stage = stages[stage_index + 1]
    target_slice_selector = lambda: select_slices(service, target_stage)
    copy_on_appearance_rules = [item for item in after_backup_rules if item.StagingOperationType.ref == STAGING_TYPE_COPY_ON_APPEARANCE]
    workflow = []
    if copy_on_appearance_rules:
        workflow += calculate_copy_on_appearance_workflow(copy_on_appearance_rules, backup_set_calculator, source_slice_selector, target_slice_selector, source_stage, target_stage) + cleanup_workflow

    move_on_cleanup_rules = [item for item in after_backup_rules if item.StagingOperationType.ref == STAGING_TYPE_MOVE_ON_CLEANUP]
    if move_on_cleanup_rules:
        workflow += calculate_move_on_cleanup_workflow(service, move_on_cleanup_rules, backup_set_calculator, cleanup_time, source_slice_selector, source_stage, target_stage) + cleanup_workflow

    return workflow


def calculate_workflow_command(service, argument):
    backup_plan, stage_index = argument.BackupPlan, argument.StageIndex.ref
    workflow = calculate_workflow(service, backup_plan, stage_index)
    min_stage_progress = round(100 * stage_index / len(argument.BackupPlan.Route.Stages))
    max_stage_progress = round(100 * (stage_index + 1) / len(argument.BackupPlan.Route.Stages))
    if len(workflow):
        step_delta = round((max_stage_progress - min_stage_progress) / len(workflow))
    commands = []
    for i, (cmd, arg) in enumerate(workflow):
        min_progress = min_stage_progress + i * step_delta
        max_progress = min_stage_progress + (i + 1) * step_delta
        commands.append(make_command_spec(cmd, arg, min_progress, max_progress))
    return make_workflow(commands)
