from . constants import *
from . slices import sort_by_creation_time
from calendar import monthrange
import datetime
import acrort

def relative_date(time, years=0, months=0):
    new_year = time.year + years + int((time.month + months) / 12)
    new_month = time.month + (months % 12)
    if new_month > 12:
        new_month = new_month % 12
    max_day = monthrange(new_year, new_month)[1]
    new_day = min(time.day, max_day)

    result = time.replace(year=new_year, month=new_month, day=new_day)
    return result


def create_abandon_time_calculator(rule):
    period_value = rule.DeleteOlderBackupsTimePeriod.Value.ref
    period_type = rule.DeleteOlderBackupsTimePeriod.PeriodType.ref
    creation_time = lambda slice: slice['creation_time']
    period_types = {
        TIME_PERIOD_YEARS: (lambda x: relative_date(creation_time(x), years=period_value)),
        TIME_PERIOD_MONTHS: (lambda x: relative_date(creation_time(x), months=period_value)),
        TIME_PERIOD_WEEKS: (lambda x: creation_time(x) + datetime.timedelta(weeks=period_value)),
        TIME_PERIOD_DAYS: (lambda x: creation_time(x) + datetime.timedelta(days=period_value)),
        TIME_PERIOD_HOURS: (lambda x: creation_time(x) + datetime.timedelta(hours=period_value)),
        TIME_PERIOD_MINUTES: (lambda x: creation_time(x) + datetime.timedelta(minutes=period_value)),
        TIME_PERIOD_SECONDS: (lambda x: creation_time(x) + datetime.timedelta(seconds=period_value)),
        TIME_PERIOD_MILLISECONDS: (lambda x: creation_time(x) + datetime.timedelta(milliseconds=period_value))
    }
    return period_types[period_type]


def check_slice_retention(slice, predecessor, abandon_time_calculators, backup_set_calculator, cleanup_time):
    backup_set = backup_set_calculator(slice, predecessor)
    if backup_set in abandon_time_calculators:
        abandon_time = abandon_time_calculators[backup_set]
        return abandon_time(slice) >= cleanup_time
    return True


def create_retention_predicate(rules, backup_set_calculator, cleanup_time):
    all_backup_set_rules = [rule for rule in rules if rule.BackupSetIndex.is_data() and rule.BackupSetIndex.type == acrort.plain.NIL]
    if len(all_backup_set_rules) > 0:
        adandon_time = create_abandon_time_calculator(all_backup_set_rules[0])
        return (lambda slice, pred: adandon_time(slice) >= cleanup_time)
    backup_set_rules = [item for item in rules if item.BackupSetIndex.is_data() and item.BackupSetIndex.type != acrort.plain.NIL]
    abandon_time_calculators = {rule.BackupSetIndex.ref: create_abandon_time_calculator(rule) for rule in backup_set_rules}
    return (lambda slice, predecessor: check_slice_retention(slice, predecessor, abandon_time_calculators, backup_set_calculator, cleanup_time))

# NEVER delete last slice, except setting 'Clean up archive > When there is insufficient space while backing up'
def preserve_last_slice(source_slices, affected_slices):
    if affected_slices and len(source_slices) == len(affected_slices):
        return affected_slices[:-1]
    return affected_slices

def select_slices_on_source(source_slice_selector):
    source_slices = source_slice_selector()
    if not source_slices:
        return []
    return sort_by_creation_time(source_slices)


def select_slices_abandoned_by_time(source_slices, retention_predicate):
    if not source_slices:
        return []
    slices_and_predecessors = [(slice, pred) for slice, pred in zip(source_slices, [None] + source_slices[:-1])]
    return [slice for slice, pred in slices_and_predecessors if not retention_predicate(slice, pred)]


def select_slices_abandoned_by_number(source_slices, backup_set_calculator, rules):
    if not source_slices:
        return []
    all_backup_set_rules = [rule for rule in rules if rule.BackupSetIndex.is_data() and rule.BackupSetIndex.type == acrort.plain.NIL]
    if len(all_backup_set_rules) > 0:
        upper_limit = all_backup_set_rules[0].BackupCountUpperLimit.ref
        if len(source_slices) <= upper_limit:
            return []
        else:
            return source_slices[:-upper_limit]
    slices_and_predecessors = zip(source_slices, [None] + source_slices[:-1])
    slice_dict = {}
    for slice, predecessor in slices_and_predecessors:
        backup_set = backup_set_calculator(slice, predecessor)
        if not backup_set in slice_dict:
            slice_dict[backup_set] = [slice]
        else:
            slice_dict[backup_set] += [slice]
    abandoned = []
    backup_set_rules = [rule for rule in rules if rule.BackupSetIndex.is_data() and rule.BackupSetIndex.type != acrort.plain.NIL]
    backup_set_rules = {rule.BackupSetIndex.ref: rule.BackupCountUpperLimit.ref for rule in backup_set_rules}
    for backup_set, slices in slice_dict.items():
        if backup_set in backup_set_rules:
            upper_limit = backup_set_rules[backup_set]
            if len(slices) > upper_limit:
                abandoned += slices[:-upper_limit]
    return sort_by_creation_time(abandoned)
