import acrort
import datetime
import acrobind.protection
from . command_ids import *
from . constants import *


def epoch_to_time(e):
    return datetime.datetime.fromtimestamp(e)


def get_default_object(service, name):
    result = [('^Is', 'string', 'Dml::TypeDescription')]
    if name is not None:
        result += [('.Typename', 'string', name)]
    results = service.dml.select(acrort.dml.ViewSpec(acrort.plain.Unit(flat=result)))
    if len(results) == 0:
        acrort.common.make_logic_error('type description for name={} cannot be found'.format(name)).throw()
    return results[0].DefaultObject


def get_slice_id(slice):
    for _, pair in slice.MetaInformation:
        if pair.first == 'slice-id':
            return pair.second.ref
    return None


def make_stage_spec(backup_plan, stage):
    return {
        'location_uri': stage.LocationUri,
        'location_credentials': stage.LocationCredentials,
        'archive_name': backup_plan.Route.Stages[0].ArchiveName,
        'archive_password': backup_plan.Options.BackupOptions.ArchiveProtection.Password,
        'rules': stage.Configuration.StagingRulesSet,
        'slicing': stage.ArchiveSlicing
    }


def make_slice_spec(slice):
    return {
        'kind': slice.Kind.ref,
        'creation_time': epoch_to_time(slice.CreationTime.ref),
        'id': get_slice_id(slice),
        'uri': slice.SliceUri.ref
    }


def select_slices(service, stage):
    default_archive_access = get_default_object(service, 'LocationView::ArchiveByNameAccess')
    archive_access_patch = [
        ('.ArchiveName', stage['archive_name']),
        ('.LocationUri', stage['location_uri']),
        ('.EncryptionPassword', stage['archive_password']),
        ('.LocationCredentials', stage['location_credentials'])
    ]
    archive_access = default_archive_access.merge(acrort.plain.UnitDiff(archive_access_patch))
    default_slice_selection = get_default_object(service, 'LocationView::SliceSelection')
    slice_selection_patch = [('.Archive', archive_access)]
    spec = default_slice_selection.merge(acrort.plain.UnitDiff(slice_selection_patch))

    try:
        activity_id = service.tol.launch_command(command=DMS_SELECT_SLICES_COMMAND_ID, argument=spec)
        response = service.tol.get_result(activity_id)
        return [make_slice_spec(slice) for _, slice in response.Slices]
    except acrort.Exception as e:
        failed_to_find_archive = e.to_error().find_suberror(0xa100d4) is not None
        cannot_find_archive = e.to_error().find_suberror(0xa10007) is not None
        unknown_archive = e.to_error().find_suberror(0x270030) is not None
        if failed_to_find_archive or cannot_find_archive or unknown_archive:
            return None
        raise


def select_last_slice(service, stage):
    slices = select_slices(service, stage)
    if slices is None or len(slices) == 0:
        return None
    return slices[-1]


def sort_by_creation_time(slices):
    return sorted(slices, key=lambda x: x['creation_time'], reverse=False)
