"""
@file
@brief    Certificate Storage implementation for Astorage client
@details  Copyright (c) 2001-2014 Acronis
@author   Bulat Gaifullin (bulat.gaifullin@acronis.com)
@since    $Id: authority.py 980710 2014-11-26 08:17:40Z bgaifullin $
"""

from acronis.tools import http
from xmlrpc.client import dumps as xml_dumps, loads as xml_loads
import logging

__all__ = ["CertificateAuthority"]


class CertificateAuthority:
    def get_redirect(self, account_id, subaccount_id=None):
        """
        :param account_id: unique identifier of account
        :param subaccount_id: unique identifier of subaccount, may be None,
        :return: tuple (host, cipher)
        """
        raise NotImplementedError

    def issue(self, account_id, subaccount_id=None):
        """
        :param account_id: unique identifier of account
        :param subaccount_id: unique identifier of subaccount, may be None
        :return: bytes, certificate + private key in PEM format
        """
        raise NotImplementedError

    def get_storage(self, account_id):
        """
        :param account_id: unique identifier of account
        :return: tuple(storage_address, brand_id)
        """
        return self.request('GetDatacenterAddressWithBrand', account_id)[:2]

    def create_certificate(self, request, account_id, subaccount_id):
        """
        :param request: bytes, certificate request in ASN1 format
        :param account_id: unique identifier of account
        :param subaccount_id: unique identifier of subaccount, may be None
        :return: bytes, authority signed certificate
        """
        # some clients does not have crypto library
        return self.request('CreateCertificate', account_id, request, subaccount_id=subaccount_id)[0].data

    def invalidate(self, reason, account_id, subaccount_id=None):
        """
        :param reason: certificate renew reason one of ("expired", "revoked", "unknown_ca", "quota_exceeded")
        :param account_id: unique identifier of account
        :param subaccount_id: unique identifier of machine(AKA subaccount)
        """
        pass

    def request(self, method, account_id, *args, subaccount_id=None):
        host, cipher = self.get_redirect(account_id, subaccount_id=subaccount_id)
        try:
            return xml_loads(http.post('https://' + host, data=xml_dumps((cipher,) + args, methodname=method)))[0]
        except Exception as e:
            logging.exception('{0}:{1} failed: {2}.'.format(host, method, e))
            raise
