"""
@file
@brief    
@details  Copyright (c) 2001-2013 Acronis
@author   Bulat Gaifullin (bulat.gaifullin@acronis.com)
@since    $$Id: outbox.py 971584 2014-10-30 14:05:39Z georgy.kirichenko $$
"""

from acronis.lib.algorithms import for_each
from email import encoders
from email.mime.base import MIMEBase
from email.mime.text import MIMEText
from email.mime.multipart import MIMEMultipart
import smtplib
import os


SMTPException = smtplib.SMTPException


def _make_attachment(file):
    if isinstance(file, dict):
        path = file['path']
        alias = file['alias']
        ctype = file.get('type', 'application/octet-stream')
    else:
        path = file
        alias = os.path.basename(file)
        ctype = 'application/octet-stream'

    main_type, _, sub_type = ctype.partition('/')
    attachment = MIMEBase(main_type, sub_type)
    with open(path, 'rb') as stream:
        attachment.set_payload(stream.read())
    encoders.encode_base64(attachment)
    attachment.add_header('Content-Disposition', 'attachment', filename=alias)
    return attachment


class Message(MIMEMultipart):
    def __init__(self, sender, subject, message, recipients, files=None, bodytype='plain', hidden_recipients=None):
        super().__init__()
        self.attach(MIMEText(message, bodytype))
        self['Subject'] = subject or ''
        self['From'] = sender or ''
        self['To'] = ','.join(recipients) if recipients else None
        self['Bcc'] = ','.join(hidden_recipients) if hidden_recipients else None
        if files is not None:
            for_each(files, lambda x: self.attach(_make_attachment(x)))


class SmtpClient:
    def __init__(self, host=None, port=0, username=None, password=None, encryption_algo=None, timeout=None):
        encryption_algo = encryption_algo.upper() if encryption_algo else ''
        smtp_class = smtplib.SMTP_SSL if encryption_algo in ('SSL',) else smtplib.SMTP
        encryption_algo = encryption_algo.upper()
        self.smtp = smtp_class(host=host or '127.0.0.1', port=port, timeout=timeout or 5)
        if encryption_algo == 'TLS':
            self.smtp.starttls()
        if username and password:
            self.smtp.login(username, password)

    def __enter__(self):
        return self.smtp

    def __exit__(self, *_):
        try:
            self.smtp.quit()
        except smtplib.SMTPServerDisconnected:
            pass
        return False


class SmtpConnection:
    def __init__(self, **kwargs):
        self.arguments = kwargs

    def connect(self):
        return SmtpClient(**self.arguments)


class Outbox:
    def __init__(self, connection=None, try_count=5):
        self.connection = connection or SmtpConnection()
        self.try_count = try_count

    def send(self, message):
        attempt = 0
        while True:
            try:
                with self.connection.connect() as connection:
                    connection.send_message(message)
                break
            except smtplib.SMTPServerDisconnected:
                if attempt > self.try_count:
                    raise
                attempt += 1
