# @copyright (c) 2002-2015 Acronis International GmbH. All rights reserved.
# since    $Id: backup_plans_create.py 1115299 2015-09-08 09:46:15Z atyuryumov $

__author__ = "Bulat Gaifullin (bulat.gaifullin@acronis.com)"

from acronis.tools.http import Session, urljoin, HTTPError
from copy import deepcopy
import glob
import json
import os
import sys
import time
import urllib
import uuid

RETRIES_COUNT = 2


def load_plan_templates(plan):
    result = []
    plan_files = []
    if plan is not None:
        plan_files = [os.path.join(os.path.dirname(__file__), "plans", plan),]
    else:
        plan_files = glob.glob(os.path.join(os.path.dirname(__file__), "plans", "*.json"))

    for iter in plan_files:
        with open(iter) as template:
            result.append(json.load(template))
    return result


def parse_machines_list(file_name):
    if file_name is None:
        return None
    hosts_file_path = os.path.join(os.path.dirname(__file__), file_name)
    if not os.path.isfile(hosts_file_path):
        print("File '{}' not found".format(file_name), file=sys.stderr)
        return None

    with open(hosts_file_path) as hosts_file:
        return [line.strip().lower() for line in hosts_file]


def get_plans(session, connection_info):
    plans = session.get("{}/api/ams/backup/plans".format(connection_info['host']))
    result = {}
    for plan in plans['data']:
        if plan.get('origin') == 'centralized':
            result[plan['name']] = plan
    return result


def merge_inclusions(current_inclusions, new_inclusions, exclusions):
    result = deepcopy(current_inclusions)
    for iter in exclusions:
        if iter in result:
            result.remove(iter)
            
    for iter in new_inclusions:
        if iter not in result:
            result.append(iter)
    return result


def get_machines(group_id, machines_includes, machines_excludes, session, connection_info):
    inclusions = []
    exclusions = []
    resources = []
    if not group_id:
        resources = session.get("{}/api/ams/resources".format(connection_info['host'], connection_info['host_name']))
    else:
        resources = session.get("{}/api/ams/resources/{}/children".format(connection_info['host'], connection_info['host_name'], group_id))

    for resource in resources['data']:
        if resource['type'] == 'machine':
            title = resource['title'].lower()
            if ((machines_excludes is not None) and (title in machines_excludes)):
                exclusions.append({'resource_key': resource['id'], 'key': resource['id']})
            elif ((machines_includes is None) or (title in machines_includes)):
                inclusions.append({'resource_key': resource['id'], 'key': resource['id']})
        elif resource['type'] == 'group':
            incl, excl = get_machines(resource['id'], machines_includes, machines_excludes, session, connection_info)
            inclusions.extend(incl)
            exclusions.extend(excl)
        
    return inclusions, exclusions


def create_backup_plan(plan_templates, machines_includes, machines_excludes, session, connection_info, retry):
    try:
        inclusions, exclusions = get_machines(None, machines_includes, machines_excludes, session, connection_info)
        if len(inclusions) == 0:
            return

        plans = get_plans(session, connection_info)
        for template in plan_templates:
            current_plan = plans.get(template['name'], None)
            plan = deepcopy(template)
            if current_plan:
                plan['id'] = current_plan['id']
                #plan['target']['inclusions'] = merge_inclusions(current_plan['target']['inclusions'], inclusions, exclusions)
                plan['target']['inclusions'] = current_plan['target']['inclusions']
                #result = session.post("{}/api/ams/{}/bplans/{}".format(connection_info['host'], connection_info['host_name'], current_plan['id']), data={'action': 'update', 'data': plan})
                result = session.put("{}/api/ams/backup/plans/{}".format(connection_info['host'], current_plan['id']), data=plan)
            else:
                plan['target']['inclusions'] = inclusions
                plan['id'] = str(uuid.uuid1())
                #session.post("{}/api/ams/{}/bplans".format(connection_info['host'], connection_info['host_name']), data={'action': 'create', 'data': plan})
                result = session.post("{}/api/ams/backup/plans".format(connection_info['host']), data=plan)
    except urllib.error.HTTPError as err:
        error_message = err.read()
        print(error_message)
        print(str(err))
        if retry == 0 or err.code <= 500:
            raise err
        time.sleep(5)
        print("ERROR: create_backup_plan retry", file=sys.stderr)
        create_backup_plan(plan_templates, machines_includes, machines_excludes, session, connection_info, retry - 1)


def get_backup_console(session, url, try_count=100):
    try:
        return session.get(url)
    except HTTPError as e:
        if e.code == 449:
            if try_count > 0:
                return get_backup_console(session, url, try_count - 1)
        raise


def process_group(session, hostname, group, plan_templates, machines_includes, machines_excludes, ident=0):
    print("  " * ident, ">", group["name"], file=sys.stderr, end="")
    if group["kind"] == 40:
         start_time = time.time()
         backup_console = get_backup_console(session, 'https://%s/api/1/groups/%s/backupconsole' % (hostname, group['id']))
         wcs_session = Session()
         wcs_session.host = backup_console['host']
 
         try:
             info = wcs_session.post(urljoin(backup_console['host'], 'api/remote_connection'), data={'access_token': backup_console['token']})
             info = json.loads(info.decode('utf8'))
             info['host'] = backup_console['host']
             info['host_name'] = info['hostName']
             create_backup_plan(plan_templates, machines_includes, machines_excludes, wcs_session, info, RETRIES_COUNT)
         finally:
             # do not use urljoin() - it has a bug: it deletes '/bc/' at the end of host value
             #wcs_session.get(urljoin(backup_console['host'], '/api/remote_connection/logout'))             
             wcs_session.get('{}api/remote_connection/logout'.format(backup_console['host']))
        
         print(' - done. ({}s)'.format(round(time.time() - start_time, 1)), file=sys.stderr)
    elif group["kind"] < 40 and group["kind"] > 0:
        print(' - look into', file=sys.stderr)
        groups = session.get('https://%s/api/1/groups/%s/children' % (hostname, group["id"]))
        for group in groups['items']:
            process_group(session, hostname, group, plan_templates, machines_includes, machines_excludes, ident + 1)
    else:
         print(' - skip', file=sys.stderr)


def main(args):
    session = Session()
    plan_templates = load_plan_templates(args.plan)
    machines_includes = parse_machines_list(args.inclusions)
    machines_excludes = parse_machines_list(args.exclusions)
    session.post('https://%s/api/1/login' % args.hostname, data={'username': args.username, 'password': args.password})
    group = session.get('https://%s/api/1/groups/%s' % (args.hostname, args.parent))
    process_group(session, args.hostname, group, plan_templates, machines_includes, machines_excludes)


def parse_arguments():
    from argparse import ArgumentParser

    parser = ArgumentParser(add_help=True)
    parser.add_argument('--hostname', help='hostname', default='eu2-cloud.acronis.com')
    parser.add_argument('--username', help='account name')
    parser.add_argument('--password', help='account password')
    parser.add_argument('--parent', help='parent group id')
    parser.add_argument('--plan', default=None, help='plan template file name')
    parser.add_argument('--inclusions', default=None, help='file with machines to include')
    parser.add_argument('--exclusions', default=None, help='file with machines to exclude')

    return parser.parse_args()


if __name__ == '__main__':
    main(parse_arguments())
